using System;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace Damage
{
    [SkipSerialisation]
    public class HumanDamageBehaviour : MonoBehaviour
    {
        [SkipSerialisation]
        public ShockBehaviour shockBehaviour;

        [SkipSerialisation]
        private LimbBehaviour limb;
        [SkipSerialisation]
        private PersonBehaviour person;

        [SkipSerialisation]
        public Organ[] organs;
        [SkipSerialisation]
        public Organ[] lungs;

        [SkipSerialisation]
        public bool hasOrgan;
        [SkipSerialisation]
        public float boneHealth = 50f;
        [SkipSerialisation]
        public float initialBoneHealth;
        [SkipSerialisation]
        public float boneSize = 0.8f;
        [SkipSerialisation]
        public float boneStrength = 2f;
        [SkipSerialisation]
        public float initialBoneStrength;
        [SkipSerialisation]
        public float fractureChance = 0.25f;
        [SkipSerialisation]
        public float tissueBlood;
        public static float tissueBloodMax = 5.6f; // 2 liters

        [SkipSerialisation]
        public float tissueDamage;
        [SkipSerialisation]
        public float tissueDamageThreshold;
        [SkipSerialisation]
        public float tissueDamageMultiplier;
        float shotHeat;
        [SkipSerialisation]
        public bool vitalOrganDestroyed;
        [SkipSerialisation]
        public float strength; // basestrength, so limbs can be returned to normal

        [SkipSerialisation]
        public int bulletCount;

        public float infection; // goes to 1. results from damage to the stomach or intestines
        public float infectionResistance;

        bool detailMode;

        GameObject[] detailObjects;

        private ContactPoint2D[] contactBuffer = new ContactPoint2D[8];

        bool CanRevive()
        {
            foreach (LimbBehaviour limb in person.Limbs)
            {
                HumanDamageBehaviour damage = limb.GetComponent<HumanDamageBehaviour>();
                if (damage.hasOrgan)
                {
                    foreach (Organ organ in damage.organs)
                    {
                        if (organ.vital && organ.destroyed)
                        {
                            return false;
                        }
                    }
                }
            }
            return true;
        }
        public bool CanReanimate()
        {
            // is this any faster or is this just for showing off?
            if (Array.Find(person.Limbs, x => x.gameObject.name == "Head").GetComponent<HumanDamageBehaviour>().organs[0].integrity <= 0)
                return false;
            else
                return true;
        }
        void Awake()
        {
            Global.main.LimbStatusToggled += new EventHandler<bool>(Main_LimbStatusToggled);
        }
        void Main_LimbStatusToggled(object sender, bool e)
        {
            detailMode = e;
            if (e)
            {
                if (detailObjects != null)
                {
                    foreach(GameObject obj in detailObjects)
                    {
                        obj.SetActive(true);
                    }
                }
            }
            else
            {
                SpriteRenderer sr = GetComponent<SpriteRenderer>();
                Color color = sr.color;
                color.r = 1;
                color.g = 1;
                color.b = 1;
                sr.color = color;
                if (detailObjects != null)
                {
                    foreach (GameObject obj in detailObjects)
                    {
                        obj.SetActive(false);
                    }
                }
            }
        }
        void OnDestroy()
        {
            Global.main.LimbStatusToggled -= new EventHandler<bool>(Main_LimbStatusToggled);
        }

        void Start()
        {
            limb = gameObject.GetComponent<LimbBehaviour>();
            person = limb.Person;
            strength = limb.BaseStrength;
            float pixel = ModAPI.PixelSize;

            switch (gameObject.name)
            {
                case "Head":
                    Organ brain = ScriptableObject.CreateInstance<Organ>();
                    brain.name = "Brain";
                    brain.health = 20f;
                    brain.initialHealth = 20f;
                    brain.integrity = 25f;
                    brain.initialIntegrity = 25f;
                    brain.vital = true;
                    brain.chance = .98f;
                    brain.damageMultiplier = 3f;
                    brain.organDepth = 0f;

                    organs = new Organ[]{
                        brain
                    };
                    boneSize = 1f;
                    boneHealth = 75f;
                    boneStrength = 5f;
                    fractureChance = .4f;
                    hasOrgan = true;

                    detailObjects = new GameObject[]
                    {
                        CreateDetailObject(Mod.SpriteAssets.brain, new Vector3(0, pixel), brain)
                    };
                    break;
                case "UpperBody":
                    Organ heart = ScriptableObject.CreateInstance<Organ>();
                    heart.name = "Heart";
                    heart.health = 40f;
                    heart.initialHealth = 40f;
                    heart.integrity = 25f;
                    heart.initialIntegrity = 25f;
                    heart.bleedRateMultiplier = 2f;
                    heart.vital = true;
                    heart.chance = .75f;
                    heart.organDepth = 4f;
                    heart.bleedRateMultiplier = 2f;

                    Organ rlung = ScriptableObject.CreateInstance<Organ>();
                    rlung.name = "Right Lung";
                    rlung.group = "Lungs";
                    rlung.health = 50f;
                    rlung.initialHealth = 50f;
                    rlung.integrity = 50f;
                    rlung.initialIntegrity = 50f;
                    rlung.chance = .375f;
                    rlung.organDepth = 1f;

                    Organ llung = ScriptableObject.CreateInstance<Organ>();
                    llung.name = "Left Lung";
                    llung.group = "Lungs";
                    llung.health = 50f;
                    llung.initialHealth = 50f;
                    llung.integrity = 50f;
                    llung.initialIntegrity = 50f;
                    llung.chance = .375f;
                    llung.organDepth = 1f;

                    organs = new Organ[]{
                        heart,
                        rlung,
                        llung
                    };
                    lungs = new Organ[]{
                        rlung,
                        llung
                    };
                    /*LimbBehaviour middle = limb.ConnectedLimbs.Find(x => x.gameObject.name=="MiddleBody");
                    HumanDamageBehaviour otherDamage = middle.GetComponent<HumanDamageBehaviour>();
                    if (otherDamage.organs == null)
                    {
                        otherDamage.organs = new Organ[4]{
                            rlung,
                            llung,
                            null,
                            null
                        };
                    }
                    else
                    {   
                        otherDamage.organs[2] = rlung;
                        otherDamage.organs[3] = llung;
                    }*/
                    boneSize = 9f;
                    boneStrength = 4f;
                    hasOrgan = true;

                    detailObjects = new GameObject[]
                    {
                        CreateDetailObject(Mod.SpriteAssets.heart, new Vector3(pixel, pixel * -1.5f), heart, 1),
                        CreateDetailObject(Mod.SpriteAssets.lung, new Vector3(pixel * 0.5f, -pixel), null, 0, lungs)
                    };
                    break;
                case "MiddleBody":
                    Organ stomach = ScriptableObject.CreateInstance<Organ>();
                    stomach.name = "Stomach";
                    stomach.health = 20f;
                    stomach.initialHealth = 20f;
                    stomach.integrity = 10f;
                    stomach.initialIntegrity = 10f;
                    stomach.chance = .5f;
                    stomach.organDepth = 4f;

                    Organ liver = ScriptableObject.CreateInstance<Organ>();
                    liver.name = "Liver";
                    liver.health = 50f;
                    liver.initialHealth = 50f;
                    liver.integrity = 20f;
                    liver.initialIntegrity = 20f;
                    liver.bleedRateMultiplier = 2.5f;
                    liver.chance = .75f;
                    liver.organDepth = 4f;

                    boneSize = .5f;
                    boneStrength = 4f;
                    boneHealth = 30f;

                    organs = new Organ[2]
                    {
                        stomach,
                        liver
                    };
                    /*if (organs == null)
                    {   
                        organs = new Organ[4]{
                            stomach,
                            liver,
                            null,
                            null
                        };
                    }
                    else
                    {
                        organs[2] = stomach;
                        organs[3] = liver;
                    }*/
                    hasOrgan = true;

                    detailObjects = new GameObject[]
                    {
                        CreateDetailObject(Mod.SpriteAssets.stomach, new Vector3(pixel, 0), stomach),
                        CreateDetailObject(Mod.SpriteAssets.liver, new Vector3(pixel * 0.5f, pixel * 2.5f), liver)
                    };
                    break;
                case "LowerBody":
                    Organ intestines = ScriptableObject.CreateInstance<Organ>();
                    intestines.name = "Intestines";
                    intestines.health = 75f;
                    intestines.initialHealth = 75f;
                    intestines.integrity = 60f;
                    intestines.initialIntegrity = 60f;
                    intestines.bleedRateMultiplier = 1.25f;
                    intestines.chance = .9f;
                    intestines.organDepth = 0f;

                    organs = new Organ[]{
                        intestines
                    };
                    hasOrgan = true;
                    boneSize = 0f;
                    boneStrength = 0f;

                    detailObjects = new GameObject[]
                    {
                        CreateDetailObject(Mod.SpriteAssets.intestines, new Vector3(0, 4 * pixel), intestines)
                    };
                    break;
                case "UpperLegFront":
                case "UpperLeg":
                    boneSize = 0.9f;
                    fractureChance = 0.3f;
                    boneHealth = 250f;
                    boneStrength = 8f;
                    break;
            }
            switch (limb.RoughClassification)
            {
                case LimbBehaviour.BodyPart.Arms:
                    tissueDamageThreshold = 250f;
                    tissueDamageMultiplier = 0.1f;
                    break;
                case LimbBehaviour.BodyPart.Legs:
                    tissueDamageThreshold = 250f;
                    tissueDamageMultiplier = 0.1f;
                    break;
                case LimbBehaviour.BodyPart.Head:
                    tissueDamageThreshold = 500f;
                    tissueDamageMultiplier = 0.1f;
                    break;
                case LimbBehaviour.BodyPart.Torso:
                    tissueDamageThreshold = 800f;
                    tissueDamageMultiplier = 0.1f;
                    break;
            }
            initialBoneHealth = boneHealth;
            initialBoneStrength = boneStrength;
            if (limb.IsZombie && hasOrgan)
            {
                foreach (Organ organ in organs)
                {
                    organ.integrity = Mathf.Min(organ.initialIntegrity * 0.5f, organ.integrity);
                    if (organ.name == "Brain")
                    {
                        organ.initialHealth = Mathf.Max(organ.initialHealth, 40f);
                        organ.health = organ.initialHealth;
                        organ.initialIntegrity = Mathf.Max(organ.initialIntegrity, 40f);
                        organ.integrity = organ.initialIntegrity;
                        organ.damageMultiplier = Mathf.Min(UnityEngine.Random.Range(0.1f, 0.9f), organ.damageMultiplier);
                    }
                }
            }
            GameObject CreateDetailObject(Sprite sprite, Vector3 position, Organ organ = null, int sort = 0, Organ[] organs = null) // add optout component
            {
                GameObject obj = new GameObject();
                SpriteRenderer sr = obj.AddComponent<SpriteRenderer>();
                sr.sprite = sprite;
                sr.sortingLayerName = "Top";
                sr.sortingOrder = GetComponent<SpriteRenderer>().sortingOrder + sort;
                obj.transform.parent = transform;
                obj.transform.localPosition = position;
                obj.transform.localRotation = Quaternion.Euler(0, 0, 0);
                obj.transform.localScale = Vector3.one;
                obj.AddComponent<Optout>();
                OrganViewBehaviour view = obj.AddComponent<OrganViewBehaviour>();
                view.sr = sr;
                if (organ)
                    view.organ = organ;
                if (organs != null)
                    view.lungs = organs;
                obj.SetActive(false);
                return obj;
            }
            limb.HasLungs = false;
        }
        void Update()
        {
            limb.CirculationBehaviour.InternalBleedingIntensity = 0;
            if (limb.IsZombie && limb.Person.ShockLevel > 0.0f || limb.IsZombie && limb.Person.PainLevel > 0.0f)
            {
                limb.Person.ShockLevel = 0.0f;
                limb.Person.PainLevel = 0.0f;
            }
            if (limb.CirculationBehaviour.TotalLiquidAmount > 1.40129846432482E-45)
            {
                foreach (KeyValuePair<Liquid, BloodContainer.RefFloat> keyValuePair in limb.CirculationBehaviour.LiquidDistribution)
                {
                    Liquid liquid = keyValuePair.Key;
                    string liquidID = Liquid.GetIdentity(liquid);
                    switch (liquidID)
                    {
                        case "INSTANT DEATH POISON":
                            KillBody();
                            break;
                        case "MENDING SERUM":
                            HealBone();
                            if (hasOrgan)
                            {
                                foreach (Organ organ in organs)
                                {
                                    RepairIntegrity(organ);
                                }
                            }
                            tissueBlood = 0;
                            tissueDamage = 0;
                            break;
                        case "COAGULATION SERUM":
                            if (hasOrgan)
                            {
                                foreach (Organ organ in organs)
                                {
                                    organ.bleedRate = 0;
                                }
                            }
                            break;
                    }
                }
            }

            if (hasOrgan)
            {
                foreach (Organ organ in organs)
                {
                    if (!organ.destroyed)
                    {
                        if (organ.regenerationSpeed > 0)
                        {
                            if (organ.health < organ.initialHealth)
                                organ.health += (organ.regenerationSpeed * limb.CirculationBehaviour.BloodFlow) * Time.deltaTime;
                            if (organ.integrity < organ.initialIntegrity)
                                organ.integrity += (organ.regenerationSpeed * limb.CirculationBehaviour.BloodFlow) * Time.deltaTime;
                        }
                    }
                    if (organ.vital)
                    {
                        limb.Health = limb.InitialHealth * (organ.health / organ.initialHealth);
                        if (!limb.IsZombie || limb.IsZombie && organ.name == "Brain")
                        {
                            if (!organ.destroyed)
                            {
                                if (limb.IsDismembered || limb.PhysicalBehaviour.isDisintegrated)
                                    KillBody();
                                if (limb.IsLethalToBreak && limb.Broken && !limb.IsZombie)
                                    KillBody();
                            }
                            if (organ.destroyed || vitalOrganDestroyed)
                            {
                                limb.CirculationBehaviour.IsPump = false;
                                if (!vitalOrganDestroyed)
                                    KillBody();
                                vitalOrganDestroyed = true;
                            }
                            else
                            {
                                limb.CirculationBehaviour.IsPump = true;
                                vitalOrganDestroyed = false;
                            }
                        }
                        else if (!vitalOrganDestroyed)
                        {
                            limb.CirculationBehaviour.IsPump = true;
                            vitalOrganDestroyed = false;
                        }
                    }
                }
                if (gameObject.name == "UpperBody" && !limb.IsZombie)
                {
                    Organ rightLung = lungs[0];
                    Organ leftLung = lungs[1];
                    float combinedHealth = rightLung.health + leftLung.health;
                    float combinedInitialHealth = rightLung.initialHealth + leftLung.initialHealth;
                    if (combinedHealth < combinedInitialHealth)
                    {
                        float lowerClamp = combinedHealth / combinedInitialHealth;
                        if (person.OxygenLevel > lowerClamp)
                        {
                            // cancel out the regeneration of oxygen
                            // just copied straight from dotpeek lol
                            person.OxygenLevel -= (float)(Mathf.Pow(Mathf.Sin(Time.time * 3f), 16f) * 0.5 / 90.0 * (Time.deltaTime * 60.0));
                            // remove oxygen
                            person.OxygenLevel -= Time.deltaTime / Mathf.Clamp(360 * (combinedHealth / combinedInitialHealth), 60, 360);
                        }
                    }
                }
            }
            if (!limb.PhysicalBehaviour.isDisintegrated && detailMode)
            {
                SpriteRenderer sr = GetComponent<SpriteRenderer>();
                Color color = sr.color;
                color.g = 1 - (tissueBlood / (tissueBloodMax * 1.2f)); // so that it stops at 0.2
                color.b = 1 - (tissueBlood / (tissueBloodMax * 1.2f));
                sr.color = color;
            }
        }
        void FixedUpdate()
        {
            shotHeat = 1f;
            if (hasOrgan)
            {
                foreach (Organ organ in organs)
                {
                    if (organ.integrity < organ.initialIntegrity && !limb.IsZombie)
                    {
                        float bleedRate = organ.bleedRate;
                        float integrityPercent = organ.integrity / organ.initialIntegrity;
                        bleedRate = Time.deltaTime / Mathf.Lerp(60f, 20f, limb.CirculationBehaviour.BloodFlow) * bleedRate;
                        bleedRate *= 1 - integrityPercent; // more damaged = more bleeding
                        if (limb.CirculationBehaviour.BloodFlow <= 0) // improve this later
                            bleedRate = 0;
                        if (tissueBlood < tissueBloodMax)
                        {
                            tissueBlood += bleedRate;
                        }
                        else
                        {
                            bleedRate /= limb.NodeBehaviour.Connections.Length;
                            foreach (ConnectedNodeBehaviour otherLimb in limb.NodeBehaviour.Connections)
                            {
                                if (otherLimb == null)
                                    continue;
                                HumanDamageBehaviour otherDamage = otherLimb.GetComponent<HumanDamageBehaviour>();
                                if (otherDamage.tissueBlood < tissueBloodMax)
                                {
                                    otherDamage.tissueBlood += bleedRate;
                                    otherDamage.tissueBlood = Mathf.Clamp(otherDamage.tissueBlood, 0, tissueBloodMax);
                                }
                            }
                        }
                        if (organ.name == "Stomach" || organ.name == "Intestines")
                        {
                            if (UnityEngine.Random.value < 0.01f * Time.fixedDeltaTime * (1 - integrityPercent)) // 1% chance every second
                            {
                                infection += 0.05f;
                            }
                        }
                        limb.CirculationBehaviour.Drain(bleedRate);
                        tissueBlood = Mathf.Clamp(tissueBlood, 0, tissueBloodMax);
                    }
                    if (!organ.destroyed)
                    {
                        if (limb.InternalTemperature >= 100.0)
                            organ.bleedRate = 0;
                        if (!limb.IsZombie)
                        {
                            if (tissueBlood > tissueBloodMax * 0.7f)
                            {
                                float mult = tissueBlood / (tissueBloodMax * 0.7f);
                                organ.Damage(Time.deltaTime * mult);
                                if (organ.group == "Lungs" && UnityEngine.Random.value < 0.001f * mult && organ.health / organ.initialHealth < 0.75f)
                                {
                                    CollapseLung(organ);
                                }
                                //if (organ.destroyed)
                                    //Debug.Log(organ.name + " destroyed by internal bleeding.");
                            }
                            if (infection > 0.5f)
                            {
                                organ.Damage(5f * Time.fixedDeltaTime * ((infection - 0.5f) * 2));
                            }
                            if (limb.CirculationBehaviour.BloodFlow <= 0.1f)
                            {
                                organ.Damage(0.05f);
                                //if (organ.destroyed)
                                    //Debug.Log(organ.name + " destroyed by lack of blood flow.");
                            }
                            if (person.OxygenLevel <= 0.0500000007450581)
                            {
                                organ.Damage(0.5f);
                                //if (organ.destroyed)
                                    //Debug.Log(organ.name + " destroyed by lack of oxygen.");
                            }
                            if (limb.PhysicalBehaviour.Charge > 1.40129846432482E-45)
                                organ.Damage(limb.PhysicalBehaviour.Charge * 0.01f);
                        }
                        if (limb.InternalTemperature >= limb.DiscomfortingHeatTemperature || limb.InternalTemperature <= limb.BodyTemperature - 20)
                            organ.Damage(1f / 400f);
                        if (limb.InternalTemperature <= limb.FreezingTemperature)
                            organ.Damage(40f * Time.fixedDeltaTime);
                        if (limb.PhysicalBehaviour.Temperature > limb.PhysicalBehaviour.Properties.BurningTemperatureThreshold)
                        {
                            organ.Damage((float)(0.00150000001303852 + (limb.PhysicalBehaviour.Temperature - limb.PhysicalBehaviour.Properties.BurningTemperatureThreshold) / 1000.0));
                        }
                    }
                }
            }
            if (infection > 0)
            {
                float rate = 1f / 240f * Time.fixedDeltaTime;
                float resist = 1f / 180f * Time.fixedDeltaTime * infectionResistance;
                if (infection == 1f)
                {
                    foreach (ConnectedNodeBehaviour otherLimb in limb.NodeBehaviour.Connections)
                    {
                        if (otherLimb == null)
                            continue;
                        HumanDamageBehaviour otherDamage = otherLimb.GetComponent<HumanDamageBehaviour>();
                        if (UnityEngine.Random.value < 0.05f * Time.fixedDeltaTime) // 5% chance every second
                        {
                            otherDamage.infection += 0.01f;
                        }
                    }
                }
                else if (infection < 1)
                    infection += rate;
                infection -= resist;
                infection = Mathf.Clamp01(infection);
                limb.SkinMaterialHandler.RottenProgress = Mathf.Max(limb.SkinMaterialHandler.RottenProgress, infection);
                if (limb.Health > 0 && infection > 0.5f)
                {
                    limb.Health -= limb.InitialHealth / 2 * Time.fixedDeltaTime;
                    limb.Health = Mathf.Max(0, limb.Health);
                }
            }
        }
        void Use()
        {
            /*if (limb.IsZombie && gameObject.name == "Head")
            {
                RaycastHit2D hit = Physics2D.Raycast(transform.position, transform.right * transform.lossyScale.x, 0.25f * Mathf.Abs(transform.lossyScale.x));
                if (hit.collider && hit.collider.gameObject.TryGetComponent(out LimbBehaviour otherLimb))
                {
                    if (otherLimb.Person.Limbs.Contains(limb))
                        return;
                    Debug.Log("rawr");
                    otherLimb.CirculationBehaviour.Cut(hit.point, transform.right * -transform.lossyScale.x);
                    otherLimb.Person.PainLevel += 0.01f;
                    otherLimb.PhysicalBehaviour.PlayClipOnce(((IList<AudioClip>) otherLimb.PhysicalBehaviour.Properties.StabSound).PickRandom(), 0.2f);
                    if (UnityEngine.Random.value < .3f)
                    {
                        otherLimb.CirculationBehaviour.AddLiquid(Liquid.GetLiquid("REANIMATION AGENT [Advanced Damage]"), 1.4f); // 1.4 units is half a liter
                        Debug.Log(">:3");   
                    }
                }
            }*/
        }
        public void KillBody()
        {
            //Debug.Log("Death");
            foreach (LimbBehaviour limb in limb.Person.Limbs)
            {
                HumanDamageBehaviour damageBehaviour = limb.GetComponent<HumanDamageBehaviour>();
                damageBehaviour.vitalOrganDestroyed = true;
                if (damageBehaviour.hasOrgan)
                {
                    foreach (Organ organ in damageBehaviour.organs)
                    {
                        organ.health = 0;
                        organ.destroyed = true;
                    }
                }
            }
        }
        private void OnFragmentHit(float force)
        {
            float bonerng = UnityEngine.Random.value;
            if (bonerng < boneSize)
                DamageBone(force * 2f);
            if (hasOrgan)
            {
                foreach (Organ organ in organs)
                {
                    float rng = UnityEngine.Random.value;
                    organ.Damage(force * 10);
                    organ.DamageIntegrity(force * 10);
                }
            }
            //Debug.Log("Fragment hit with force of " + force);
        }
        private void Stabbed(Stabbing stab)
        {
            if (!stab.stabber.StabCausesWound)
                return;
            float artery = UnityEngine.Random.value;
            if (hasOrgan)
            {
                foreach (Organ organ in organs)
                {
                    float damage = UnityEngine.Random.Range(0, organ.initialHealth * 0.5f);
                    float rng = UnityEngine.Random.value * 0.85f;
                    if (organ.name == "Brain")
                        damage = 50f;
                    DamageOrgan(damage, organ, false, rng, true);
                }
            }
            if (limb.RoughClassification == LimbBehaviour.BodyPart.Arms && artery < 0.3f)
                HitArtery(0.5f);
            if (limb.RoughClassification == LimbBehaviour.BodyPart.Legs && artery < 0.2f)
                HitArtery(1f);
            if (limb.RoughClassification == LimbBehaviour.BodyPart.Torso && artery < 0.05f)
                HitArtery(0.25f);
            if (limb.NodeBehaviour.IsConnectedToRoot)
                shockBehaviour.EnterShock(UnityEngine.Random.value * 2);
        }
        private void Shot(Shot shot) // make less bullet power result in less damage
        {
            // damage values:
            // 0.5 bump, 1 very small, 4 small, 8 medium, 20 large, 50 massive
            shotHeat += 2;
            person.AddPain(shot.damage / 5f);
            float damage = shot.damage;

            float bulletPower = Mathf.Clamp(damage, 2f, 80f);
            float initialPower = bulletPower;
            float chanceMultiplier = 1f;

            float bonerng = UnityEngine.Random.value;
            float fragmentrng = UnityEngine.Random.value;

            if (boneHealth > 0 && bonerng < boneSize)
            {
                DamageBone(damage);
                bulletPower -= boneStrength;
            }
            if (damage > 10 && UnityEngine.Random.value < 0.75f)
            {
                damage *= UnityEngine.Random.Range(1.5f, 2f);
                //Debug.Log("Yawing");
                bulletPower = Mathf.Max(bulletPower / 2, bulletPower - 20);
                chanceMultiplier *= 0.95f;
            }
            if (fragmentrng < 0.05f)
            {
                FragmentDamage(UnityEngine.Random.Range(2, 3), Mathf.Clamp(damage / 5, 0, 25));
                //Debug.Log("Bullet fragmentation");
                bulletPower = Mathf.Max(bulletPower / 1.6f, bulletPower - 16);
            }
            //Debug.Log(bulletPower);
            damage -= shot.damage * (1 - (bulletPower / initialPower)); // so that yawing still increases damage
            if (hasOrgan)
            {
                foreach (Organ organ in organs)
                {
                    if (Mathf.Max(bulletPower, 0) < organ.organDepth)
                        continue;
                    if (organ.name == "Brain" && !limb.IsZombie)
                        person.Consciousness = 0f;
                    float organrng = UnityEngine.Random.value * chanceMultiplier;
                    DamageOrgan(damage, organ, false, organrng, true);
                }
            }
            foreach (ConnectedNodeBehaviour node in limb.NodeBehaviour.Connections)
            {
                if (!node)
                    continue;
                LimbBehaviour otherLimb = node.GetComponent<LimbBehaviour>();
                float rng = UnityEngine.Random.value;
                if (limb.Health <= 0)
                    otherLimb.Damage(shot.damage / 5);
                if (rng < 0.2f)
                {
                    if (limb.Health > 0)
                        otherLimb.Damage(shot.damage / 20);
                    if (otherLimb.RoughClassification == limb.RoughClassification && otherLimb.gameObject.GetComponent<HumanDamageBehaviour>().hasOrgan)
                    {
                        foreach (Organ organ in otherLimb.gameObject.GetComponent<HumanDamageBehaviour>().organs)
                        {
                            float organrng = UnityEngine.Random.value;
                            DamageOrgan(damage / 5, organ, false, organrng, true);
                        }
                    }
                }
            }
            if (shot.damage >= (tissueDamageThreshold - tissueDamage) * (1 - (limb.SkinMaterialHandler.RottenProgress / 1.25f)) / shotHeat)
                DamageTissue(shot.damage, 1);
            float artery = UnityEngine.Random.value;
            if (limb.RoughClassification == LimbBehaviour.BodyPart.Arms && artery < 0.3f)
                HitArtery(0.5f);
            if (limb.RoughClassification == LimbBehaviour.BodyPart.Legs && artery < 0.2f)
                HitArtery(1f);
            if (limb.RoughClassification == LimbBehaviour.BodyPart.Torso && artery < 0.025f)
                HitArtery(5f);
            if (limb.NodeBehaviour.IsConnectedToRoot)
                shockBehaviour.EnterShock(Mathf.Min(shot.damage, 10f));
            bulletCount += 1;
        }
        private void ExitShot(Shot shot)
        {
            if (shot.damage >= (tissueDamageThreshold - tissueDamage) * (1 - (limb.SkinMaterialHandler.RottenProgress / 1.25f)) / shotHeat)
                DamageTissue(shot.damage, 2);
            bulletCount -= 1;
        }
        public void DamageTissue(float damage, float multiplier)
        {
            damage *= multiplier * shotHeat * ((limb.SkinMaterialHandler.RottenProgress * 2) + 1);
            tissueDamage += damage * tissueDamageMultiplier;
            //Debug.Log("Tissue damaged by " + damage * tissueDamageMultiplier + " damage. Tissue health now at " + (tissueDamageThreshold - tissueDamage));
            if (tissueDamage >= tissueDamageThreshold)
            {
                limb.BreakingThreshold = 0.0f;
                //Debug.Log("Dismemberment");
            }
        }
        public void HitArtery(float severity)
        {
            limb.CirculationBehaviour.BleedingRate += severity;
            //Debug.Log("Artery broken");
        }
        public void DamageBone(float damage)
        {
            if (boneHealth <= 0)
                return;
            if (damage <= boneStrength || boneStrength <= 0)
            {
                limb.Damage(damage);
                boneStrength = Mathf.Max(boneStrength - (damage / 5), 0);
                return;
            }
            float random = UnityEngine.Random.value;
            // damage values:
            // 0.5 bump, 1 very small, 4 small, 8 medium, 20 large, 50 massive
            // bone strength values:
            // 0.5 shattered, 1 damaged, 2 light, 4 medium, 6 strong, 8 very strong 

            // high damage causes more fragmentation w/ more variation of fragment count
            //int fragmentCount = UnityEngine.Random.Range((damage/2)-boneStrength, damage-boneStrength);
            //fragmentCount = Mathf.Max(0, fragmentCount);
            //float totalFragDamage = 1/(boneHealth/initialBoneHealth);
            // fragmentation will NEVER do more total damage than what damaged it or than its health
            //totalFragDamage = Mathf.Clamp(totalFragDamage, 0, Mathf.Min(damage, boneHealth));

            float totalFragDamage = 0;
            float fragmentCount = 0;

            if (damage > boneStrength * 3 && random < (fractureChance / 5) * (damage / (boneStrength * 3)))
            {
                // shattering, cause a lot of fragmentation and a ton of bone damage
                limb.BaseStrength = 0;
                float boneDamage = Mathf.Max(100, damage);
                limb.Person.ShockLevel = 1f;
                limb.Person.PainLevel = 1f;

                totalFragDamage = Mathf.Min(boneDamage, boneHealth);
                fragmentCount = UnityEngine.Random.Range(Mathf.Max(boneDamage / 2f - boneStrength, 0), boneDamage - boneStrength);

                boneHealth = Mathf.Max(boneHealth - boneDamage, 0);
                boneStrength = Mathf.Min(boneStrength, 0.5f);

                //Debug.Log("Shattered bone. " + "Health at " + boneHealth);

                limb.PhysicalBehaviour.PlayClipOnce(((IList<AudioClip>)person.BoneBreakClips).PickRandom<AudioClip>());
            }
            else if (random < fractureChance)
            {
                // fracturing, cause some fragmentation and a bit of bone damage
                limb.BaseStrength /= boneStrength;
                limb.Person.ShockLevel += 0.2f;
                limb.Person.PainLevel += 0.2f;
                float boneDamage = initialBoneHealth / boneStrength;

                totalFragDamage = Mathf.Min(boneDamage, boneHealth);
                fragmentCount = UnityEngine.Random.Range(boneDamage / 3f, boneDamage);

                boneHealth -= boneDamage;
                boneStrength = Mathf.Max(boneStrength - (boneDamage / 2), 0);

                //Debug.Log("Fractured bone. " + "Health at " + boneHealth);

                limb.PhysicalBehaviour.PlayClipOnce(((IList<AudioClip>)person.BoneBreakClips).PickRandom<AudioClip>());
            }
            // if the bone is neither shattered nor fractured, cause minimal fragmentation and a small amount of bone damage
            else
            {
                float boneDamage = damage - boneStrength;
                boneHealth -= boneDamage;
                boneStrength -= boneDamage / 15;
            }
            if (fragmentCount > 0)
                FragmentDamage(fragmentCount, totalFragDamage / fragmentCount);
            boneHealth = Mathf.Max(boneHealth, 0);
        }
        void FragmentDamage(float amount, float damage)
        {
            int neighboringLimbs = limb.NodeBehaviour.Connections.Length;
            amount /= neighboringLimbs;
            foreach (ConnectedNodeBehaviour node in limb.NodeBehaviour.Connections)
            {
                if (!node)
                    continue;
                LimbBehaviour otherLimb = node.GetComponent<LimbBehaviour>();
                otherLimb.Damage(damage * amount);

                if (otherLimb.TryGetComponent(out HumanDamageBehaviour damageBehaviour))
                {
                    if (damageBehaviour.hasOrgan)
                    {
                        foreach (Organ organ in damageBehaviour.organs)
                        {
                            float rng = UnityEngine.Random.value;
                            damageBehaviour.DamageOrgan((damage * amount) * rng, organ, true, 0, true);
                        }
                    }
                }
            }
        }
        public void DamageOrgan(float damage, Organ organ, bool ignoreMultiplier, float rng, bool physical)
        {
            float rawDamage = damage;
            if (!ignoreMultiplier)
                damage *= organ.damageMultiplier;
            if (!organ.destroyed && rng < organ.chance)
            {
                organ.Damage(damage);
                if (physical)
                {
                    organ.DamageIntegrity(rawDamage);
                    if ((organ.name == "Intestines" || organ.name == "Stomach") && organ.integrity / organ.initialIntegrity <= 0.75f)
                    {
                        if (UnityEngine.Random.value <= 1f * (1 - (organ.integrity / organ.initialIntegrity)))
                            infection += 0.01f;
                    }
                    if (organ.group == "Lungs" && UnityEngine.Random.value > 0.25f && rawDamage > 6 && organ.integrity / organ.initialIntegrity < 0.8f)
                    {
                        CollapseLung(organ);
                    }
                }
                if (organ.name == "Heart" && limb.CirculationBehaviour.BleedingRate > 0f)
                {
                    limb.CirculationBehaviour.BleedingRate *= 1.5f;
                }
                //Debug.Log(organ.name + " damaged by " + damage + ". Health is now " + organ.health);
                //if (organ.destroyed)
                    //Debug.Log(organ.name + " destroyed.");
            }
        }
        public void CollapseLung(Organ organ)
        {
            if (organ.group == "Lungs")
            {
                organ.health = 0;
                organ.destroyed = true;
                limb.Person.ShockLevel = Mathf.Max(0.75f, limb.Person.ShockLevel);
                limb.Person.PainLevel = Mathf.Max(0.9f, limb.Person.PainLevel);
                //Debug.Log("Lung collapsed");
            }
        }
        public void HealBone()
        {
            //Debug.Log("Bone healed.");
            boneHealth = initialBoneHealth;
            boneStrength = initialBoneStrength;
            limb.BaseStrength = strength;
        }
        public void RepairOrgan(Organ organ)
        {
            //Debug.Log(organ.name + " repaired.");
            organ.health = organ.initialHealth;
            organ.integrity = organ.initialIntegrity;
            if (organ.vital)
            {
                if (vitalOrganDestroyed && CanRevive())
                    Revive();
                vitalOrganDestroyed = false;
            }
            organ.destroyed = false;
            organ.bleedRate = 0;
        }
        public void RepairIntegrity(Organ organ)
        {
            //Debug.Log(organ.name + " integrity repaired.");
            organ.health = organ.initialHealth;
            organ.integrity = organ.initialIntegrity;
            organ.bleedRate = 0;
        }
        public void RepairLimb()
        {
            infection = 0.0f;
            tissueDamage = 0f;
            tissueBlood = 0f;
        }
        public void Revive()
        {
            foreach (LimbBehaviour limb in person.Limbs)
            {
                HumanDamageBehaviour damage = limb.GetComponent<HumanDamageBehaviour>();
                if (damage)
                {
                    damage.vitalOrganDestroyed = false;
                }
            }
        }
        // mainly exists for compatability purposes
        // regenerates humans to full health, much like the life syringe
        public void HealBody()
        {
            foreach (LimbBehaviour limb in person.Limbs)
            {
                if (limb.TryGetComponent(out HumanDamageBehaviour damage))
                {
                    if (damage.hasOrgan)
                    {
                        damage.HealBone();
                        damage.tissueBlood = 0;
                        damage.tissueDamage = 0;
                        foreach (Organ organ in damage.organs)
                        {
                            RepairOrgan(organ);
                        }
                    }
                }
            }
        }
        [Obsolete]
        [SkipSerialisation]
        public enum Part
        {
            Head,
            UpperTorso,
            LowerTorso,
            Stomach,
        }
        void OnCollisionEnter2D(Collision2D collision)
        {
            float impulse = ((IEnumerable<ContactPoint2D>)contactBuffer).Take(collision.GetContacts(contactBuffer)).Max(c => c.normalImpulse);
            if (impulse > 1f)
            {
                float rng = UnityEngine.Random.value;
                float fracturerng = UnityEngine.Random.value;
                //Debug.Log(impulse);
                if (boneStrength > 0)
                {
                    boneHealth -= impulse;
                    impulse -= boneStrength / 10;
                    DamageBone(impulse);
                }
                if (hasOrgan)
                {
                    foreach (Organ organ in organs)
                    {
                        if (impulse > 2f)
                        {
                            DamageOrgan(impulse * 5, organ, true, 0, true);
                        }
                        else
                            DamageOrgan(impulse * 5, organ, true, rng * 0.8f, false);
                    }
                }
            }
        }
    }
    public class Organ : ScriptableObject
    {
        public string name;
        // optional
        public string group;

        public float health;
        public float initialHealth;

        public float integrity;
        public float initialIntegrity;

        public float bleedRate;
        public float bleedRateMultiplier = 1f;

        public float chance;
        public float damageMultiplier = 1f;
        public bool destroyed;
        public bool vital;
        public float regenerationSpeed = 0f;

        // just for bullets right now
        public float organDepth;

        public void Damage(float damage)
        {
            health -= damage;
            if (health <= 0)
                destroyed = true;
            health = Mathf.Clamp(health, 0, initialHealth);
        }
        public void DamageIntegrity(float damage, bool bleed = true)
        {
            integrity -= damage;
            integrity = Mathf.Clamp(integrity, 0, initialIntegrity);
            //bleedRate += damage / 24f * bleedRateMultiplier; // 1/4 of external bleeding rate
            if (bleed)
            {
                float max = 10 * bleedRateMultiplier;
                float percent = Mathf.Clamp01(damage / 24f * bleedRateMultiplier / max);
                bleedRate += Mathf.SmoothStep(0, max, percent);
            }

            //Debug.Log("Bleeding rate at " + bleedRate);
        }
    }

    public class ShockBehaviour : MonoBehaviour
    {
        public PersonBehaviour person;
        float initialShockTime;
        float shockTime;

        void Update()
        {
            if (shockTime > 0)
            {
                person.AdrenalineLevel = 1;
                person.ShockLevel = Mathf.Min(person.ShockLevel, 1f * (1 - Mathf.Min(shockTime / initialShockTime, 0.5f)));
                person.PainLevel = Mathf.Min(person.PainLevel, 1f * (1 - Mathf.Min(shockTime / initialShockTime, 0.5f)));
                shockTime = Mathf.Max(shockTime - Time.deltaTime, 0f);
                if (shockTime == 0)
                    initialShockTime = 0;
            }
        }
        public void EnterShock(float time)
        {
            shockTime = Mathf.Min(shockTime + time, 10f);
            initialShockTime += time;
            //Debug.Log("In shock for " + shockTime);
        }
    }
    public class OrganViewBehaviour : MonoBehaviour 
    {
        public Organ organ;
        public Organ[] lungs;
        public SpriteRenderer sr;
        void Update()
        {
            if (lungs != null)
            {
                float percent = (lungs[0].health + lungs[1].health) / (lungs[0].initialHealth + lungs[1].initialHealth);
                Color color = sr.color;
                color.r = 1;
                color.b = percent;
                color.g = percent;
                color.a = 0.9f;
                if (percent == 0)
                {
                    color.r = 0.25f;
                    color.a = 0.6f;
                }
                sr.color = color;
                return;
            }
            if (organ.health != organ.initialHealth)
            {
                Color color = sr.color;
                color.r = 1;
                color.b = organ.health / organ.initialHealth;
                color.g = organ.health / organ.initialHealth;
                color.a = 0.9f;
                if (organ.health == 0)
                {
                    color.r = 0.25f;
                    color.a = 0.6f;
                }
                sr.color = color;
            }
            else
            {
                Color color = sr.color;
                color.r = 1;
                color.b = 1;
                color.g = 1;
                color.a = 0.9f;
                sr.color = color;
            }
        }
    }
}