using UnityEngine;

namespace Damage
{
    public class Mod
    {
        public static bool spawnerEnabled = true;
        public static bool zombieBites = true;
        public static bool restrictedReconstructor = false;
        public static PhysicalProperties humanProperties;
        public struct SpriteAssets
        {
            public static Sprite brain;
            public static Sprite lung;
            public static Sprite heart;
            public static Sprite stomach;
            public static Sprite liver;
            public static Sprite intestines;

            public static Sprite hemostat;
        }
        public static void Main()
        {
            ModAPI.RegisterLiquid(UltraStrengthOrganSyringe.UltraStrengthOrganSerumLiquid.ID, new UltraStrengthOrganSyringe.UltraStrengthOrganSerumLiquid());
            ModAPI.RegisterLiquid(ZombieOrganSyringe.ZombiePoisonOrganLiquid.ID, new ZombieOrganSyringe.ZombiePoisonOrganLiquid());
            ModAPI.RegisterLiquid(OrganLifeSyringe.OrganLifeSerumLiquid.ID, new OrganLifeSyringe.OrganLifeSerumLiquid());
            ModAPI.RegisterLiquid(ClottingSyringe.ClottingSerum.ID, new ClottingSyringe.ClottingSerum());
            ModAPI.RegisterLiquid(RepairSyringe.RepairSerum.ID, new RepairSyringe.RepairSerum());
            ModAPI.RegisterLiquid(Antibiotic.ID, new Antibiotic());
            ModAPI.Register<SpawnBehaviour>();
            LimbStatusViewBehaviour.Main.gameObject.AddComponent<InspectBehaviour>();

            SpriteAssets.brain = ModAPI.LoadSprite("sprites/brain.png");
            SpriteAssets.lung = ModAPI.LoadSprite("sprites/lungs.png");
            SpriteAssets.heart = ModAPI.LoadSprite("sprites/heart.png");
            SpriteAssets.stomach = ModAPI.LoadSprite("sprites/stomach.png");
            SpriteAssets.intestines = ModAPI.LoadSprite("sprites/intestines.png");
            SpriteAssets.liver = ModAPI.LoadSprite("sprites/liver.png");
            SpriteAssets.hemostat = ModAPI.LoadSprite("sprites/hemostat.png");

            ModAPI.OnItemSpawned += (sender, obj) =>
            {
                GameObject gameObj = obj.Instance;
                PersonBehaviour person = gameObj.GetComponent<PersonBehaviour>();
                if (spawnerEnabled)
                {
                    if (person)
                    {
                        if (person.Limbs[0].SpeciesIdentity == "Human")
                        {
                            GameObject shock = new GameObject();
                            shock.transform.parent = obj.Instance.transform;
                            shock.AddComponent<Optout>();
                            ShockBehaviour shockBehaviour = shock.AddComponent<ShockBehaviour>();
                            shockBehaviour.person = person;
                            foreach (LimbBehaviour limbBehaviour in person.Limbs)
                            {
                                if (!limbBehaviour.gameObject.GetComponent<HumanDamageBehaviour>())
                                {
                                    limbBehaviour.gameObject.AddComponent<HumanDamageBehaviour>();
                                    limbBehaviour.PhysicalBehaviour.Properties = humanProperties;
                                }
                                limbBehaviour.GetComponent<HumanDamageBehaviour>().shockBehaviour = shockBehaviour;
                            }
                        }
                        else if (person.Limbs[0].SpeciesIdentity == "Android")
                        {
                            foreach (LimbBehaviour limbBehaviour in person.Limbs)
                            {
                                if (!limbBehaviour.gameObject.GetComponent<AndroidDamageBehaviour>())
                                    limbBehaviour.gameObject.AddComponent<AndroidDamageBehaviour>();
                            }
                        }
                    }
                    if (gameObj.TryGetComponent(out HealthGunBehaviour health))
                    {
                        HealthGunOrganBehaviour gun = obj.Instance.AddComponent<HealthGunOrganBehaviour>();
                        gun.range = health.range;
                        gun.offset = health.offset;
                        gun.PhysicalBehaviour = health.PhysicalBehaviour;
                        gun.ShotSounds = health.ShotSounds;
                        gun.Muzzleflash = health.Muzzleflash;
                        gun.barrelPosition = health.barrelPosition;
                        gun.barrelDirection = health.barrelDirection;
                        Object.Destroy(health);
                    }
                    if (gameObj.TryGetComponent(out UltraStrengthSyringe strength))
                    {
                        UltraStrengthOrganSyringe syringe = obj.Instance.AddComponent<UltraStrengthOrganSyringe>();
                        Object.Destroy(strength);
                    }
                    if (gameObj.TryGetComponent(out ZombieSyringe zombie))
                    {
                        ZombieOrganSyringe syringe = obj.Instance.AddComponent<ZombieOrganSyringe>();
                        Object.Destroy(zombie);
                    }
                    if (gameObj.TryGetComponent(out LifeSyringe life))
                    {
                        OrganLifeSyringe syringe = obj.Instance.AddComponent<OrganLifeSyringe>();
                        Object.Destroy(life);
                    }
                }
            };
            ModAPI.Register(
                new Modification()
                {
                    OriginalItem = ModAPI.FindSpawnable("Acid Syringe"),
                    NameOverride = "Clotting Syringe [Advanced Damage]",
                    DescriptionOverride = "Syringe full of a serum that slows down blood loss.",
                    CategoryOverride = ModAPI.FindCategory("Biohazard"),
                    ThumbnailOverride = ModAPI.LoadSprite("clotthumbnail.png"),
                    NameToOrderByOverride = "Syringe2",
                    AfterSpawn = (Instance) =>
                    {
                        UnityEngine.Object.Destroy(Instance.GetComponent<SyringeBehaviour>());
                        Instance.GetOrAddComponent<ClottingSyringe>();
                    }
                }
            );
            ModAPI.Register(
                new Modification()
                {
                    OriginalItem = ModAPI.FindSpawnable("Acid Syringe"),
                    NameOverride = "Repair Syringe [Advanced Damage]",
                    DescriptionOverride = "Syringe full of a serum which repairs organs.",
                    CategoryOverride = ModAPI.FindCategory("Biohazard"),
                    ThumbnailOverride = ModAPI.LoadSprite("repairthumbnail.png"),
                    NameToOrderByOverride = "Syringe1",
                    AfterSpawn = (Instance) =>
                    {
                        UnityEngine.Object.Destroy(Instance.GetComponent<SyringeBehaviour>());
                        Instance.GetOrAddComponent<RepairSyringe>();
                    }
                }
            );
            ModAPI.Register(
                new Modification()
                {
                    OriginalItem = ModAPI.FindSpawnable("Stick"),
                    NameOverride = "Hemostat [Advanced Damage]",
                    DescriptionOverride = "Hemostatic dressing which stops bleeding.",
                    CategoryOverride = ModAPI.FindCategory("Misc."),
                    ThumbnailOverride = ModAPI.LoadSprite("hemostat view.png"),
                    AfterSpawn = (Instance) =>
                    {
                        Instance.GetOrAddComponent<HemostatBehaviour>();

                        Instance.GetComponent<SpriteRenderer>().sprite = SpriteAssets.hemostat;
                        Instance.FixColliders();
                    }
                }
            );
            ModAPI.Register(
                new Modification()
                {
                    OriginalItem = ModAPI.FindSpawnable("Blood Flask"),
                    NameOverride = "Antibiotics [Advanced Damage]",
                    DescriptionOverride = "1L Flask full of a liquid that can remove infections.",
                    CategoryOverride = ModAPI.FindCategory("Biohazard"),
                    NameToOrderByOverride = "Oil Flask",
                    ThumbnailOverride = ModAPI.LoadSprite("antibiotic view.png"),
                    AfterSpawn = (Instance) =>
                    {
                        Instance.GetComponent<FlaskBehaviour>().StartLiquid = new BloodContainer.SerialisableDistribution
                        {
                            LiquidID = Antibiotic.ID,
                            Amount = 2.8f //a flask can only hold 2.8 units of liquid (1 L)
                        };
                    }
                }
            );
        }
    }
}