using System.Collections;
using UnityEngine;
using UnityEngine.Events;
using System.Linq;
using System;
using TMPro;

namespace Damage
{
    public class SpawnBehaviour : MonoBehaviour
    {
        public bool spawnerEnabled = true;
        int i;
        
        void Start()
        {
            ModAPI.Notify("Advanced Damage added! Double press M to enable or disable applying the script to spawned humans.");
            StartCoroutine("DoublePress");

            PhysicalProperties prop = Instantiate(ModAPI.FindPhysicalProperties("Human"));

            prop.BulletSpeedAbsorptionPower = 1f;

            Mod.humanProperties = prop;
        }
        void Update()
        {
            if (Input.GetKeyDown(KeyCode.M))
            {
                i++;
                if (i >= 2)
                {
                    if (Mod.spawnerEnabled)
                    {
                        ModAPI.Notify("Automatic applying disabled!");
                        Mod.spawnerEnabled = false;
                    }
                    else
                    {
                        ModAPI.Notify("Automatic applying enabled!");
                        Mod.spawnerEnabled = true;
                    }
                    i = 0;
                }
            }
        }
        private IEnumerator DoublePress()
        {
            while (true)
            {
                i = 0;
                yield return new WaitForSeconds(1f);
            }
        }
    }
    public class InspectBehaviour : MonoBehaviour
    {
        public float t = 0.0f;
        void Update()
        {
            /*if (!LimbStatusViewBehaviour.Main.gameObject.activeSelf)
            {
                t = 0.0f;
                return;
            }*/
            t += Time.unscaledDeltaTime;
            if (t > LimbStatusViewBehaviour.Main.UpdateInterval)
            {
                TextMeshProUGUI stats = LimbStatusViewBehaviour.Main.LimbDamageSourceStats;
                t = 0.0f;
                bool bleeding = false;
                bool leaking = false;
                float infection = 0;
                int total = 0;
                stats.text += "<color=#FF5555>";
                foreach (LimbBehaviour limb in LimbStatusViewBehaviour.Main.Limbs)
                {
                    if (limb.TryGetComponent(out HumanDamageBehaviour damage))
                    {
                        if (damage.hasOrgan)
                        {
                            foreach (Organ organ in damage.organs)
                            {
                                string name = organ.name;
                                if (organ.name == "Right Lung" || organ.name == "Left Lung")
                                    continue;
                                if (organ.integrity == 0)
                                    stats.text += name + " destroyed\n";
                                else if (organ.destroyed)
                                    stats.text += name + " dead\n";
                                else if (organ.integrity < organ.initialIntegrity)
                                    stats.text += name + " damaged\n";
                                if (organ.bleedRate > 0)
                                    bleeding = true;
                            }
                            if (limb.gameObject.name == "UpperBody")
                            {
                                float integrity = (damage.lungs[0].integrity + damage.lungs[1].integrity) / 2;
                                bool destroyed = damage.lungs[0].destroyed && damage.lungs[1].destroyed;
                                bool damaged = (damage.lungs[0].integrity < damage.lungs[0].initialIntegrity) || (damage.lungs[1].integrity < damage.lungs[1].initialIntegrity);

                                if (integrity == 0)
                                    stats.text += "lungs destroyed\n";
                                else if (destroyed)
                                    stats.text += "lungs dead\n";
                                else if (damaged)
                                    stats.text += "lungs damaged\n";
                                if ((damage.lungs[0].bleedRate > 0) || (damage.lungs[1].bleedRate > 0))
                                    bleeding = true;
                            }
                        }
                        if (damage.boneHealth / damage.initialBoneHealth <= 0.35f)
                            stats.text += "shattered bone\n";
                        else if (damage.boneHealth < damage.initialBoneHealth)
                            stats.text += "fractured bone\n";
                        infection += damage.infection;
                        total++;
                    }
                    if (limb.TryGetComponent(out AndroidDamageBehaviour android))
                    {
                        if (android.hasOrgan)
                        {
                            foreach (AndroidOrgan organ in android.organs)
                            {
                                string name = organ.name;
                                if (organ.integrity == 0)
                                    stats.text += name + " destroyed\n";
                                else if (organ.destroyed)
                                    stats.text += name + " disabled\n";
                                else if (organ.integrity < organ.initialIntegrity)
                                    stats.text += name + " damaged\n";
                                if (organ.leakRate > 0)
                                    leaking = true;
                            }
                        }
                        if (limb.BaseStrength < android.initialStrength)
                            stats.text += "motor damaged\n";
                    }
                }
                if (bleeding)
                    stats.text += "internal bleeding\n";
                if (leaking)
                    stats.text += "leaking\n";
                stats.text += "</color>";
                if (infection > 0)
                    stats.text += "<color=#FF5555>" + Mathf.RoundToInt(infection / total * 100) + "% infected</color>";
            }
        }
    }
}