using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace Damage
{
    [SkipSerialisation]
    public class UltraStrengthOrganSyringe : SyringeBehaviour
    {
        public override string GetLiquidID() => UltraStrengthOrganSerumLiquid.ID;

        public class UltraStrengthOrganSerumLiquid : TemporaryBodyLiquid
        {
            public const string ID = "ULTRA STRENGTH SERUM [Advanced Damage]";

            public UltraStrengthOrganSerumLiquid()
            {
                Color = new Color(0.7604228f, 0.7137255f, 1f);
            }

            public override void OnEnterContainer(BloodContainer container)
            {
            }

            public override void OnEnterLimb(LimbBehaviour limb)
            {
                if (limb.SpeciesIdentity == "Android")
                    return;
                limb.Vitality *= 0.1f;
                limb.RegenerationSpeed += 5f;
                limb.ImpactPainMultiplier *= 0.1f;
                limb.InitialHealth += 150f;
                limb.Health = limb.InitialHealth;
                limb.BreakingThreshold += 10f;
                limb.BaseStrength = Mathf.Min(15f, limb.BaseStrength + 5f);
                limb.ShotDamageMultiplier *= 0.1f;
                limb.CirculationBehaviour.HealBleeding();
                if (limb.gameObject.TryGetComponent(out HumanDamageBehaviour damage))
                {
                    damage.strength = limb.BaseStrength;
                    damage.tissueDamageThreshold += 150f;
                    damage.tissueDamage = 0;
                    damage.initialBoneHealth += 150f;
                    damage.boneHealth = damage.initialBoneHealth;
                    damage.fractureChance *= 0.95f;
                    damage.initialBoneStrength *= 1.1f;
                    damage.boneStrength = damage.initialBoneStrength;
                    damage.infectionResistance += 0.25f;
                    damage.infectionResistance = Mathf.Min(damage.infectionResistance, 2);
                    if (damage.hasOrgan)
                    {
                        foreach (Organ organ in damage.organs)
                        {
                            organ.initialHealth += 20f;
                            organ.health = organ.initialHealth;
                            organ.initialIntegrity += 25f;
                            organ.integrity = organ.initialIntegrity;
                            organ.damageMultiplier *= 0.95f;
                            organ.regenerationSpeed += 1f;
                        }
                    }
                }
                limb.Wince(10f);
            }

            public override void OnExitContainer(BloodContainer container)
            {
            }
        }
    }
    [SkipSerialisation]
    public class ZombieOrganSyringe : SyringeBehaviour
    {
        public override string GetLiquidID() => ZombiePoisonOrganLiquid.ID;

        public class ZombiePoisonOrganLiquid : Liquid
        {
            public const string ID = "REANIMATION AGENT [Advanced Damage]";

            public ZombiePoisonOrganLiquid()
            {
                Color = new Color(0.8835177f, 1f, 0.7122642f);
            }

            public override void OnEnterContainer(BloodContainer container)
            {
            }

            public override void OnEnterLimb(LimbBehaviour limb)
            {
                if (limb.SpeciesIdentity == "Android")
                    return;
                limb.IsZombie = true;
                limb.Health = limb.InitialHealth;
                limb.Numbness = 0.0f;
                CirculationBehaviour circulationBehaviour = limb.CirculationBehaviour;
                float amount = circulationBehaviour.GetAmount(limb.GetOriginalBloodType());
                if (amount < 0.100000001490116)
                {
                    double num = circulationBehaviour.AddLiquid(limb.GetOriginalBloodType(), 1f - amount);
                }
                circulationBehaviour.BloodFlow = 1f;
                circulationBehaviour.HealBleeding();
                circulationBehaviour.BleedingRate = 0.0f;
                circulationBehaviour.IsPump = circulationBehaviour.WasInitiallyPumping;
                if (limb.gameObject.TryGetComponent(out HumanDamageBehaviour damage))
                {
                    if (!damage.CanReanimate())
                        return;
                    damage.Revive();
                    damage.vitalOrganDestroyed = false;
                    if (damage.hasOrgan)
                    {
                        foreach (Organ organ in damage.organs)
                        {
                            organ.destroyed = false;
                            organ.integrity = Mathf.Min(organ.initialIntegrity * 0.5f, organ.integrity);
                            damage.Revive();
                            if (organ.name == "Brain")
                            {
                                organ.initialHealth = Mathf.Max(organ.initialHealth, 40f);
                                organ.health = organ.initialHealth;
                                organ.initialIntegrity = Mathf.Max(organ.initialIntegrity, 40f);
                                organ.chance *= UnityEngine.Random.Range(0.35f, 0.9f);
                                organ.integrity = organ.initialIntegrity;
                                organ.damageMultiplier = Mathf.Min(UnityEngine.Random.Range(0.1f, 0.9f), organ.damageMultiplier);
                            }
                        }
                    }
                }
                if (limb.RoughClassification != LimbBehaviour.BodyPart.Head)
                    return;
                limb.Person.Consciousness = 1f;
                limb.Person.ShockLevel = 0.0f;
                limb.Person.PainLevel = 0.0f;
                limb.Person.OxygenLevel = 1f;
                limb.Person.AdrenalineLevel = 1f;
            }

            public override void OnExitContainer(BloodContainer container)
            {
            }
        }
    }
    public class OrganLifeSyringe : SyringeBehaviour
    {
        public override string GetLiquidID() => OrganLifeSerumLiquid.ID;

        public class OrganLifeSerumLiquid : TemporaryBodyLiquid
        {
            public const string ID = "LIFE SERUM [Advanced Damage]";

            public OrganLifeSerumLiquid()
            {
                Color = new Color(0.6745283f, 1f, 0.7830077f);
            }

            public override void OnEnterContainer(BloodContainer container)
            {
            }

            public override void OnEnterLimb(LimbBehaviour limb)
            {
                if (limb.SpeciesIdentity == "Android")
                    return;
                GeneralHealing(limb);
                limb.CirculationBehaviour.StartCoroutine(HealBurn(limb));
                limb.CirculationBehaviour.StartCoroutine(HealBulletHoles(limb));
                limb.CirculationBehaviour.StartCoroutine(HealWounds(limb));
                limb.CirculationBehaviour.StartCoroutine(HealRottenFlesh(limb));
                limb.CirculationBehaviour.StartCoroutine(HealAcidFlesh(limb));
                limb.LungsPunctured = false;
            }

            public override void OnExitContainer(BloodContainer container)
            {
                container.StopCoroutine("HealBurn");
                container.StopCoroutine("HealBulletHoles");
                container.StopCoroutine("HealWounds");
                container.StopCoroutine("HealRottenFlesh");
                container.StopCoroutine("HealAcidFlesh");
            }

            private void GeneralHealing(LimbBehaviour limb)
            {
                limb.HealBone();
                limb.Health = limb.InitialHealth;
                limb.Numbness = 0.0f;
                limb.CirculationBehaviour.HealBleeding();
                limb.CirculationBehaviour.IsPump = limb.CirculationBehaviour.WasInitiallyPumping;
                limb.CirculationBehaviour.BloodFlow = 1f;
                double num = limb.CirculationBehaviour.AddLiquid(limb.GetOriginalBloodType(), Mathf.Max(0.0f, 1f - limb.CirculationBehaviour.GetAmount(limb.GetOriginalBloodType())));
                limb.BruiseCount = 0;
                limb.CirculationBehaviour.GunshotWoundCount = 0;
                limb.CirculationBehaviour.StabWoundCount = 0;
                if (limb.TryGetComponent(out HumanDamageBehaviour damage))
                {
                    damage.HealBone();
                    if (damage.hasOrgan)
                    {
                        foreach (Organ organ in damage.organs)
                        {
                            damage.RepairOrgan(organ);
                        }
                    }
                    damage.RepairLimb();
                }
                if (limb.RoughClassification != LimbBehaviour.BodyPart.Head)
                    return;
                limb.Person.Consciousness = 1f;
                limb.Person.ShockLevel = 0.0f;
                limb.Person.PainLevel = 0.0f;
                limb.Person.OxygenLevel = 1f;
                limb.Person.AdrenalineLevel = 1f;
            }

            private IEnumerator HealBurn(LimbBehaviour limb)
            {
                OrganLifeSerumLiquid lifeSerumLiquid = this;
                limb.PhysicalBehaviour.Temperature = limb.BodyTemperature;
                while (limb.PhysicalBehaviour.BurnProgress > 0.100000001490116)
                {
                    limb.PhysicalBehaviour.BurnProgress -= Time.deltaTime * 0.16f;
                    yield return new WaitForEndOfFrame();
                    if (limb.CirculationBehaviour.GetAmount(lifeSerumLiquid) <= 1.40129846432482E-45)
                        yield break;
                }
                limb.PhysicalBehaviour.BurnProgress = 0.0f;
            }

            private IEnumerator HealAcidFlesh(LimbBehaviour limb)
            {
                OrganLifeSerumLiquid lifeSerumLiquid = this;
                while (limb.SkinMaterialHandler.AcidProgress > 0.100000001490116)
                {
                    limb.SkinMaterialHandler.AcidProgress -= Time.deltaTime * 0.8f;
                    yield return new WaitForEndOfFrame();
                    if (limb.CirculationBehaviour.GetAmount(lifeSerumLiquid) <= 1.40129846432482E-45)
                        yield break;
                }
                limb.SkinMaterialHandler.AcidProgress = 0.0f;
            }

            private IEnumerator HealRottenFlesh(LimbBehaviour limb)
            {
                OrganLifeSerumLiquid lifeSerumLiquid = this;
                while (limb.SkinMaterialHandler.RottenProgress > 0.100000001490116)
                {
                    limb.SkinMaterialHandler.RottenProgress -= Time.deltaTime * 0.8f;
                    yield return new WaitForEndOfFrame();
                    if (limb.CirculationBehaviour.GetAmount(lifeSerumLiquid) <= 1.40129846432482E-45)
                        yield break;
                }
                limb.SkinMaterialHandler.RottenProgress = 0.0f;
            }

            private IEnumerator HealBulletHoles(LimbBehaviour limb)
            {
                OrganLifeSerumLiquid lifeSerumLiquid = this;
                while (((IEnumerable<Vector4>)limb.SkinMaterialHandler.bulletHolePoints).Min<Vector4>(c => c.z) < 50.0)
                {
                    for (int index = 0; index < limb.SkinMaterialHandler.bulletHolePoints.Length; ++index)
                    {
                        limb.SkinMaterialHandler.bulletHolePoints[index].z = Mathf.Max(limb.SkinMaterialHandler.bulletHolePoints[index].z, 0.05f);
                        limb.SkinMaterialHandler.bulletHolePoints[index].z += Time.deltaTime * 0.5f;
                    }
                    limb.SkinMaterialHandler.Sync();
                    yield return new WaitForEndOfFrame();
                    if (limb.CirculationBehaviour.GetAmount(lifeSerumLiquid) <= 1.40129846432482E-45)
                        break;
                }
            }

            private IEnumerator HealWounds(LimbBehaviour limb)
            {
                OrganLifeSerumLiquid lifeSerumLiquid = this;
                while (((IEnumerable<Vector4>)limb.SkinMaterialHandler.damagePoints).Min<Vector4>(c => c.z) < 50.0)
                {
                    for (int index = 0; index < limb.SkinMaterialHandler.damagePoints.Length; ++index)
                        limb.SkinMaterialHandler.damagePoints[index].z += Time.deltaTime * 0.5f;
                    limb.SkinMaterialHandler.Sync();
                    yield return new WaitForEndOfFrame();
                    if (limb.CirculationBehaviour.GetAmount(lifeSerumLiquid) <= 1.40129846432482E-45)
                        break;
                }
            }
        }
    }
    public class ClottingSyringe : SyringeBehaviour
    {
        public override string GetLiquidID() => ClottingSerum.ID;
        public class ClottingSerum : TemporaryBodyLiquid
        {
            public const string ID = "Clotting Serum [Advanced Damage]";
            public override float RemovalChancePerSecond
            {
                get
                {
                    return base.RemovalChancePerSecond * 10f;
                }
            }
            public override bool ShouldCallOnEnterEveryUpdate => false;

            public ClottingSerum()
            {
                Color = new Color(0.48f, 0.57f, 0.34f, 1);
            }
            public override void OnEnterLimb(LimbBehaviour limb)
            {
                if (limb.CirculationBehaviour.BleedingRate > 1)
                {
                    float bleed = limb.CirculationBehaviour.BleedingRate;
                    bleed *= (1 - Mathf.Pow(0.6f, bleed)); // reduces bleeding rate inversely, for graph it is y = x(1 - a^x)
                    limb.CirculationBehaviour.BleedingRate = Mathf.Max(bleed, 1);
                }
                if (limb.TryGetComponent(out HumanDamageBehaviour damage) && damage.hasOrgan)
                {
                    foreach (Organ organ in damage.organs)
                    {
                        organ.bleedRate *= (1 - Mathf.Pow(0.5f, organ.bleedRate)); // modify a for this one - bleeding externally is faster than internally
                    }
                }
            }

            //Called every second by every container for every liquid it contains.
            public override void OnUpdate(BloodContainer container)
            {
                //for syringe liquids, this is usually empty.
                //if you want a syringe effect to be continuously applied, you should implement the effects here instead.
                //make sure to check if the given container is actually a CirculationBehaviour.
                base.OnUpdate(container);
                //do note that this has to call base.OnUpdate(container) if you are derving from TemporaryBodyLiquid
            }

            //Called when this liquid enters a container. Limbs are also containers. 
            //Note that this may be called quite often for the same container as liquid quickly moves in and out of it.
            public override void OnEnterContainer(BloodContainer container)
            {
                //for syringe liquids, this is usually empty.
            }

            //Called when this liquid exits a container. 
            //Note that this may be called quite often for the same container as liquid quickly moves in and out of it.
            public override void OnExitContainer(BloodContainer container)
            {
                //for syringe liquids, this is usually empty.
            }
        }
    }
    public class RepairSyringe : SyringeBehaviour
    {
        public override string GetLiquidID() => RepairSerum.ID;
        public class RepairSerum : TemporaryBodyLiquid
        {
            public const string ID = "Repair Serum [Advanced Damage]";
            public override float RemovalChancePerSecond
            {
                get
                {
                    return base.RemovalChancePerSecond * 10f;
                }
            }

            public RepairSerum()
            {
                Color = new Color(0.41f, 0.69f, 0.61f, 1);
            }
            public override void OnEnterLimb(LimbBehaviour limb)
            {
                if (limb.TryGetComponent(out HumanDamageBehaviour damage) && damage.hasOrgan)
                {
                    foreach (Organ organ in damage.organs)
                    {
                        damage.RepairOrgan(organ);
                    }
                }
            }

            //Called every second by every container for every liquid it contains.
            public override void OnUpdate(BloodContainer container)
            {
                //for syringe liquids, this is usually empty.
                //if you want a syringe effect to be continuously applied, you should implement the effects here instead.
                //make sure to check if the given container is actually a CirculationBehaviour.
                base.OnUpdate(container);
                //do note that this has to call base.OnUpdate(container) if you are derving from TemporaryBodyLiquid
            }

            //Called when this liquid enters a container. Limbs are also containers. 
            //Note that this may be called quite often for the same container as liquid quickly moves in and out of it.
            public override void OnEnterContainer(BloodContainer container)
            {
                //for syringe liquids, this is usually empty.
            }

            //Called when this liquid exits a container. 
            //Note that this may be called quite often for the same container as liquid quickly moves in and out of it.
            public override void OnExitContainer(BloodContainer container)
            {
                //for syringe liquids, this is usually empty.
            }
        }
    }
    public class Antibiotic : TemporaryBodyLiquid
    {
        public const string ID = "Antibiotic [Advanced Damage]";
        public override float RemovalChancePerSecond
        {
            get
            {
                return base.RemovalChancePerSecond * 0.5f;
            }
        }
        public override bool ShouldCallOnEnterEveryUpdate
        {
            get
            {
                return true;
            }
        }
        public Antibiotic()
        {
            Color = new Color(1f, 0.99f, 0.94f, 1);
        }
        public override void OnEnterLimb(LimbBehaviour limb)
        {
            if (limb.TryGetComponent(out HumanDamageBehaviour damage) && damage.infection > 0)
            {
                damage.infection -= 0.05f;
                damage.infection = Mathf.Clamp01(damage.infection);
            }
        }

        //Called every second by every container for every liquid it contains.
        public override void OnUpdate(BloodContainer container)
        {
            //for syringe liquids, this is usually empty.
            //if you want a syringe effect to be continuously applied, you should implement the effects here instead.
            //make sure to check if the given container is actually a CirculationBehaviour.
            base.OnUpdate(container);
            //do note that this has to call base.OnUpdate(container) if you are derving from TemporaryBodyLiquid
        }

        //Called when this liquid enters a container. Limbs are also containers. 
        //Note that this may be called quite often for the same container as liquid quickly moves in and out of it.
        public override void OnEnterContainer(BloodContainer container)
        {
            //for syringe liquids, this is usually empty.
        }

        //Called when this liquid exits a container. 
        //Note that this may be called quite often for the same container as liquid quickly moves in and out of it.
        public override void OnExitContainer(BloodContainer container)
        {
            //for syringe liquids, this is usually empty.
        }
    }
}